local Bat = Class{__includes = require 'modules.entity'}
local Explosion = require 'modules.explosion'
local EnemyAttack = require 'modules.enemyattack'
-----------------------------------------------------------
-- Init functions
-----------------------------------------------------------
function Bat:init(x,y )
  self.pos = Vector(x,y)
  self.isBat = true

  self.w = 8
  self.h = 8
  self.zIndex = 6

  self.speed = 20
  self.defaultSpeed = 20
  self.hp = 2

  self.flammability = 0.8
  self.isMobile = true
  self.dir = 0

  self:setUpAnimations()

  -- behaviors
  self.behavior = 'wander'
  self.behaviorDuration = 0.5
  self.behaviorTimer = Timer.every(self.behaviorDuration, function()
    if not Game.paused then self:chooseBehavior() end
  end)
  self:chooseRandomDirection()

  Game.world:add(self, self.pos.x, self.pos.y, self.w, self.h)
end

function Bat:setUpAnimations()
  self.anims = {}
  local w, h = SPRITESHEET:getDimensions()
  local l = 0
  local t = 112

  local g = Anim8.newGrid(8, 8, w, h, l, t)

  self.anims.spawn = Anim8.newAnimation(g(1,1, 2,1), .1)

  self.anims.walkLeft = Anim8.newAnimation(g(3,1, 4,1), .1)
  self.anims.walkRight = self.anims.walkLeft:clone():flipH()

  self.currentAnim = self.anims.spawn
  self.prevFacing = 'Left'
end

-----------------------------------------------------------
-- Update functions
-----------------------------------------------------------
function Bat:update(dt)
  local s = self.speed * dt
  local dx, dy = self.dir.x, self.dir.y
  local movementVector = Vector(dx, dy) * s
  self:move(movementVector)
  self.currentAnim:update(dt)
end

function Bat:chooseBehavior()

  local behavior = self.behavior or 'wander'

  if Game.playerKilled or self.onFire then
    behavior = 'wander'
  else
    behavior = 'approach'
  end

  self.behavior = behavior
  if self[behavior] then
    self[behavior](self)
  end
end

function Bat:move(vec)
  local oldX = self.pos.x
  local goalVec = self.pos + vec
  local filter = self.filter
  local actualX, actualY, cols, len = Game.world:move(self, goalVec.x, goalVec.y, filter)
  self.pos.x = actualX
  self.pos.y = actualY

  -- set animation

  if actualX < oldX then
    self.prevFacing = 'Left'
  elseif actualX > oldX then
    self.prevFacing = 'Right'
  end

  if self.behavior == 'attack' then
    self.currentAnim = self.anims['attack' .. self.prevFacing]
  else
    self.currentAnim = self.anims['walk' .. self.prevFacing]
  end
end

function Bat.filter(item, other)
  if other.collides == false or other.isEnemyAttack or other.isBoss then
    return false
  else
    return 'slide'
  end
end

-----------------------------------------------------------
-- Draw functions
-----------------------------------------------------------
function Bat:draw()
  local x, y = self.pos:unpack()
  -- for some reason, anim8 needs the scale...
  local s = 1
  self.currentAnim:draw(SPRITESHEET, x * s, y * s, 0, s, s)
end

-----------------------------------------------------------
-- Behavior functions
-----------------------------------------------------------
function Bat:wander()
  self.speed = self.defaultSpeed
  self:chooseRandomDirection()
end

function Bat:chooseRandomDirection()
  local rads = love.math.random(0, math.pi * 2)
  self.dir = Vector.fromPolar(rads, 1)
end

function Bat:flee()
  local roy = Game.player

  local dx = self.pos.x - roy.pos.x
  local dy = self.pos.y - roy.pos.y

  self.speed = self.defaultSpeed
  self.dir = Vector.new(dx, dy):normalizeInplace()
end

function Bat:approach()
  local roy = Game.player

  local dx = roy.pos.x - self.pos.x
  local dy = roy.pos.y - self.pos.y

  self.speed = self.defaultSpeed
  self.dir = Vector.new(dx, dy):normalizeInplace()
end

function Bat:attack()
  self.speed = 0
  self.attackTimer = Timer.after(0.3, function()
    self:spawnAttackEntity()
  end)
end

-----------------------------------------------------------
-- Gameplay functions
-----------------------------------------------------------
function Bat:spawnAttackEntity()
  local x, y = self.pos.x, self.pos.y
  local type = self.type
  local roy = Game.player

  local dx = roy.pos.x - self.pos.x
  local dy = roy.pos.y - self.pos.y

  local dir = Vector.new(dx, dy):normalizeInplace()
  table.insert(Game.entities, EnemyAttack(x, y, dir, 'default'))
end

function Bat:receiveDamage(amt, src)
  self.hp = self.hp - amt
  if self.hp <= 0 then
    self:kill()
  end
end

function Bat:kill()
  local explosionChance = .1
  if love.math.random() <= explosionChance then
    table.insert(Game.entities, Explosion(self.pos.x, self.pos.y, true))
  end

  if self.attackTimer then Timer.cancel(self.attackTimer) end
  Timer.cancel(self.behaviorTimer)
  self:remove()
end

function Bat:distanceToRoy()
  if Game.player then
    local selfMid = Vector(self.pos.x + self.w / 2, self.pos.y + self.h /2)
    local roy = Game.player
    local royMid = Vector(roy.pos.x + roy.w / 2, roy.pos.y + roy.h /2)
    return selfMid:dist(royMid)
  else
    return nil
  end
end

return Bat
