local Boss = Class{__includes = require 'modules.entity'}
local EnemyAttack = require 'modules.enemyattack'
local settings = require 'setup.settings'
local Corpse = require 'modules.corpse'
local Explosion = require 'modules.explosion'
local Bat = require 'modules.bat'
-----------------------------------------------------------
-- Init functions
-----------------------------------------------------------
function Boss:init(x,y)
  self.pos = Vector(x,y)
  self.isBoss = true

  self.w = 16
  self.h = 16
  self.zIndex = 5

  self.speed = 16
  self.defaultSpeed = 16
  self.hp = 50
  self.batChance = 0.3

  self.flammability = 0.1
  self.isMobile = true
  self.dir = 0

  self:setUpAnimations()

  -- behaviors
  self.behavior = 'wander'
  self.behaviorDuration = 0.5
  self.behaviorTimer = Timer.every(self.behaviorDuration, function()
    if not Game.paused then self:chooseBehavior() end
  end)
  self:chooseRandomDirection()

  Game.world:add(self, self.pos.x, self.pos.y, self.w, self.h)
end

function Boss:setUpAnimations()
  self.anims = {}
  local w, h = SPRITESHEET:getDimensions()
  local l = 0
  local t = 96

  local g = Anim8.newGrid(16, 16, w, h, l, t)

  self.anims.idle = Anim8.newAnimation(g(1,1), 1, 'pauseAtEnd')

  self.anims.walk = Anim8.newAnimation(g(1,1, 2,1), .2)

  self.anims.attack = Anim8.newAnimation(g(2,1, 3,1, 3,1, 1,1, 1,1), .1)

  self.anims.teleportOut = Anim8.newAnimation(g('4-8', 1), .1)

  self.anims.teleportIn = Anim8.newAnimation(g(4,1, 3, 1), .2)

  self.currentAnim = self.anims.idle
end

-----------------------------------------------------------
-- Update functions
-----------------------------------------------------------
function Boss:update(dt)
  local s = self.speed * dt
  local dx, dy = self.dir.x, self.dir.y
  local movementVector = Vector(dx, dy) * s
  self:move(movementVector)
  self.currentAnim:update(dt)
end

function Boss:chooseBehavior()
  local dist = self:distanceToRoy()
  local type = self.type
  local behavior = self.behavior or 'wander'
  local fleeDist = 48
  local teleportDist = 32
  local wanderDist = 128
  local attackDist = 72

  if Game.playerKilled or self.onFire or dist >= wanderDist then
    behavior = 'wander'
  elseif dist <= fleeDist then
    behavior = 'flee'
  elseif dist <= attackDist then
    behavior = 'attack'
  else
    behavior = 'approach'
  end

  self.behavior = behavior
  if self[behavior] then
    self[behavior](self)
  end
end

function Boss:move(vec)
  local oldX = self.pos.x
  local goalVec = self.pos + vec
  local filter = self.filter
  local actualX, actualY, cols, len = Game.world:move(self, goalVec.x, goalVec.y, filter)
  self.pos.x = actualX
  self.pos.y = actualY

  -- set animation

  if actualX < oldX then
    self.prevFacing = 'Left'
  elseif actualX > oldX then
    self.prevFacing = 'Right'
  end

  if self.behavior == 'attack' then
    self.currentAnim = self.anims.attack
  elseif self.behavior == 'teleportOut' then
    self.currentAnim = self.anims.teleportOut
  elseif self.currentBehavior == 'teleportIn' then
    self.currentAnim = self.anims.teleportIn
  else
    self.currentAnim = self.anims.walk
  end
end

function Boss.filter(item, other)
  if other.collides == false or other.isEnemyAttack or other.isBat then
    return false
  else
    return 'slide'
  end
end

-----------------------------------------------------------
-- Draw functions
-----------------------------------------------------------
function Boss:draw()
  local x, y = self.pos:unpack()
  -- for some reason, anim8 needs the scale...
  local s = 1
  self.currentAnim:draw(SPRITESHEET, x * s, y * s, 0, s, s)
end

-----------------------------------------------------------
-- Behavior functions
-----------------------------------------------------------
function Boss:wander()
  self.speed = self.defaultSpeed
  self:chooseRandomDirection()
end

function Boss:chooseRandomDirection()
  local rads = love.math.random(0, math.pi * 2)
  self.dir = Vector.fromPolar(rads, 1)
end

function Boss:flee()
  local roy = Game.player

  local dx = self.pos.x - roy.pos.x
  local dy = self.pos.y - roy.pos.y

  self.speed = self.defaultSpeed
  self.dir = Vector.new(dx, dy):normalizeInplace()
end

function Boss:approach()
  local roy = Game.player

  local dx = roy.pos.x - self.pos.x
  local dy = roy.pos.y - self.pos.y

  self.speed = self.defaultSpeed
  self.dir = Vector.new(dx, dy):normalizeInplace()
end

function Boss:attack()
  self.speed = 0
  self.attackTimer = Timer.after(0.3, function()
    if love.math.random() <= self.batChance then
      self:summonBats()
    else
      self:spawnAttackEntity()
    end
  end)
  TEsound.play({'audio/magic-1.wav', 'audio/magic-2.wav', 'audio/magic-3.wav'}, 'sfx')
end

function Boss:teleportOut()
  self.speed = 0
  Timer.after(.5, function()
    self:actuallyTeleport()
  end)
end

function Boss:teleportIn()
  self.speed = 0
end

function Boss:actuallyTeleport()

  local tx, ty
  local w, h = self.w, self.h
  local i = 1
  local maxTries = 10
  local found = false

  local minX, minY = -64, -64
  local maxX, maxY = 64, 64

  local filter = function(item)
    if item.collides == false then
      return false
    else
      return true
    end
  end

  while i < maxTries and not found do
    tx = love.math.random(minX, maxX)
    ty = love.math.random(minY, maxY)
    local items, len = Game.world:queryRect(tx, ty, w, h, filter)
    if len == 0 then
      found = true
      Game.world:update(self, tx, ty)
    else
      i = i + 1
    end
    self.behavior = 'teleportIn'
  end
end

-----------------------------------------------------------
-- Gameplay functions
-----------------------------------------------------------
function Boss:spawnAttackEntity()
  local x, y = self.pos.x, self.pos.y
  local roy = Game.player

  local dx = roy.pos.x - self.pos.x
  local dy = roy.pos.y - self.pos.y

  local dir = Vector.new(dx, dy):normalizeInplace()
  table.insert(Game.entities, EnemyAttack(x, y, dir, 'boss'))
end

function Boss:receiveDamage(amt, src)
  self.hp = self.hp - amt
  if self.hp <= 0 then
    self:kill()
  end
end

function Boss:kill()
  local explosionChance = .2
  if love.math.random() <= explosionChance then
    table.insert(Game.entities, Explosion(self.pos.x, self.pos.y, true))
  end
  table.insert(Game.entities, Corpse(self.pos.x, self.pos.y))

  if self.attackTimer then Timer.cancel(self.attackTimer) end
  Timer.cancel(self.behaviorTimer)
  Game:killedBoss()
  self:remove()
end

function Boss:distanceToRoy()
  if Game.player then
    local selfMid = Vector(self.pos.x + self.w / 2, self.pos.y + self.h /2)
    local roy = Game.player
    local royMid = Vector(roy.pos.x + roy.w / 2, roy.pos.y + roy.h /2)
    return selfMid:dist(royMid)
  else
    return nil
  end
end

function Boss:summonBats()
  local x, y = self.pos:unpack()
  table.insert(Game.entities, Bat(x, y))
end

return Boss
