local Villager = Class{__includes = require 'modules.entity'}
local Corpse = require 'modules.corpse'
local Explosion = require 'modules.explosion'
local EnemyAttack = require 'modules.enemyattack'
-----------------------------------------------------------
-- Init functions
-----------------------------------------------------------
function Villager:init(x,y, type)
  self.pos = Vector(x,y)
  self.isVillager = true

  if type then
    self.type = type
  else
    local types = {'default', 'soldier', 'archer', 'wizard'}
    local i = love.math.random(1,#types)
    self.type = types[i]
  end

  self.w = 8
  self.h = 8
  self.quad = lg.newQuad(0, 24, 8, 8, SPRITESHEET:getDimensions())
  self.zIndex = 5

  self.speed = 16
  self.defaultSpeed = 16
  self.hp = 5

  self.flammability = 0.5
  self.isMobile = true
  self.dir = 0

  self:setUpAnimations()

  -- behaviors
  self.behavior = 'wander'
  self.behaviorDuration = 0.5
  self.behaviorTimer = Timer.every(self.behaviorDuration, function()
    if not Game.paused then self:chooseBehavior() end
  end)
  self:chooseRandomDirection()

  Game.world:add(self, self.pos.x, self.pos.y, self.w, self.h)
end

function Villager:setUpAnimations()
  self.anims = {}
  local w, h = SPRITESHEET:getDimensions()
  local l = 0
  local t = 24

  if self.type == 'soldier' then t = 32
  elseif self.type == 'archer' then t = 40
  elseif self.type == 'wizard' then t = 48
  end

  local g = Anim8.newGrid(8, 8, w, h, l, t)

  self.anims.idleLeft = Anim8.newAnimation(g(1,1), 1, 'pauseAtEnd')
  self.anims.idleRight = self.anims.idleLeft:clone():flipH()

  self.anims.walkLeft = Anim8.newAnimation(g(1,1, 2,1), .1)
  self.anims.walkRight = self.anims.walkLeft:clone():flipH()

  self.anims.attackLeft = Anim8.newAnimation(g(3,1, 3,1, 4,1, 1,1, 1,1), .1)
  self.anims.attackRight = self.anims.attackLeft:clone():flipH()

  self.currentAnim = self.anims.walkLeft
  self.prevFacing = 'Left'
end

-----------------------------------------------------------
-- Update functions
-----------------------------------------------------------
function Villager:update(dt)
  local s = self.speed * dt
  local dx, dy = self.dir.x, self.dir.y
  local movementVector = Vector(dx, dy) * s
  self:move(movementVector)
  self.currentAnim:update(dt)
end

function Villager:chooseBehavior()
  local dist = self:distanceToRoy()
  local type = self.type
  local behavior = self.behavior or 'wander'
  local fleeDist = 32
  local wanderDist = 128
  local attackDist = 16

  if type == 'archer' or type == 'wizard' then
    attackDist = 48
  end

  if Game.playerKilled or self.onFire or dist >= wanderDist then
    behavior = 'wander'
  elseif type == 'default' and dist <= fleeDist then
    behavior = 'flee'
  elseif dist <= attackDist then
    behavior = 'attack'
  else
    behavior = 'approach'
  end

  self.behavior = behavior
  if self[behavior] then
    self[behavior](self)
  end
end

function Villager:move(vec)
  local oldX = self.pos.x
  local goalVec = self.pos + vec
  local filter = self.filter
  local actualX, actualY, cols, len = Game.world:move(self, goalVec.x, goalVec.y, filter)
  self.pos.x = actualX
  self.pos.y = actualY

  -- set animation

  if actualX < oldX then
    self.prevFacing = 'Left'
  elseif actualX > oldX then
    self.prevFacing = 'Right'
  end

  if self.behavior == 'attack' then
    self.currentAnim = self.anims['attack' .. self.prevFacing]
  else
    self.currentAnim = self.anims['walk' .. self.prevFacing]
  end
end

function Villager.filter(item, other)
  if other.collides == false or other.isEnemyAttack then
    return false
  else
    return 'slide'
  end
end

-----------------------------------------------------------
-- Draw functions
-----------------------------------------------------------
function Villager:draw()
  local x, y = self.pos:unpack()
  -- for some reason, anim8 needs the scale...
  local s = 1
  self.currentAnim:draw(SPRITESHEET, x * s, y * s, 0, s, s)
end

-----------------------------------------------------------
-- Behavior functions
-----------------------------------------------------------
function Villager:wander()
  self.speed = self.defaultSpeed
  self:chooseRandomDirection()
end

function Villager:chooseRandomDirection()
  local rads = love.math.random(0, math.pi * 2)
  self.dir = Vector.fromPolar(rads, 1)
end

function Villager:flee()
  local roy = Game.player

  local dx = self.pos.x - roy.pos.x
  local dy = self.pos.y - roy.pos.y

  self.speed = self.defaultSpeed
  self.dir = Vector.new(dx, dy):normalizeInplace()
end

function Villager:approach()
  local roy = Game.player

  local dx = roy.pos.x - self.pos.x
  local dy = roy.pos.y - self.pos.y

  self.speed = self.defaultSpeed
  self.dir = Vector.new(dx, dy):normalizeInplace()
end

function Villager:attack()
  self.speed = 0
  self.attackTimer = Timer.after(0.3, function()
    self:spawnAttackEntity()
  end)
  self:playAttackSound()
end

function Villager:playAttackSound()
  local type = self.type

  if self.type == 'archer' then
    TEsound.play({'audio/arrow-1.wav', 'audio/arrow-2.wav', 'audio/arrow-3.wav'}, 'sfx')

  elseif self.type == 'wizard' then
    TEsound.play({'audio/magic-1.wav', 'audio/magic-2.wav', 'audio/magic-3.wav'}, 'sfx')

  else
    TEsound.play({'audio/sword-1.wav', 'audio/sword-2.wav'}, 'sfx')

  end

end

-----------------------------------------------------------
-- Gameplay functions
-----------------------------------------------------------
function Villager:spawnAttackEntity()
  local x, y = self.pos.x, self.pos.y
  local type = self.type
  local roy = Game.player

  local dx = roy.pos.x - self.pos.x
  local dy = roy.pos.y - self.pos.y

  local dir = Vector.new(dx, dy):normalizeInplace()
  table.insert(Game.entities, EnemyAttack(x, y, dir, self.type))
end

function Villager:receiveDamage(amt, src)
  self.hp = self.hp - amt
  if self.hp <= 0 then
    self:kill()
  end
end

function Villager:kill()
  local explosionChance = .2
  if love.math.random() <= explosionChance then
    table.insert(Game.entities, Explosion(self.pos.x, self.pos.y, true))
  end
  table.insert(Game.entities, Corpse(self.pos.x, self.pos.y))

  if self.attackTimer then Timer.cancel(self.attackTimer) end
  Timer.cancel(self.behaviorTimer)
  self:remove()
end

function Villager:distanceToRoy()
  if Game.player then
    local selfMid = Vector(self.pos.x + self.w / 2, self.pos.y + self.h /2)
    local roy = Game.player
    local royMid = Vector(roy.pos.x + roy.w / 2, roy.pos.y + roy.h /2)
    return selfMid:dist(royMid)
  else
    return nil
  end
end

return Villager
