Game = {}

local Roy = require 'modules.roy'
local Villager = require 'modules.villager'
local Cottage = require 'modules.cottage'
local Tree = require 'modules.tree'
local Water = require 'modules.water'
local Wheat = require 'modules.wheat'
local Fire = require 'modules.fire'
local Brick = require 'modules.brick'
local Door = require 'modules.door'
local Explosion = require 'modules.explosion'
local World = require 'modules.world'

local Prefab = require 'modules.prefab'
-----------------------------------------------------------
-- Init functions
-----------------------------------------------------------
function Game:init()
  self.cameraSmoother = Camera.smooth.damped(8)
  self.muted = false
end

function Game:enter()
  self:reset()
end

function Game:reset()
  Timer.clear()
  self.paused = false
  self.isGameOver = false
  self.playerKilled = false
  self.isNewGame = true
  self.isBossDead = false
  Timer.after(3, function()
    self.isNewGame = false
  end)

  self.entities = {}
  self.world = Bump.newWorld()

  local worldSetup = World(1280, 720)
  self.camera = Camera(self.player.pos.x, self.player.pos.y, 2, 0)
  TEsound.playLooping('audio/screaming-skill.mp3', 'bgm', nil, .5)
end

function Game:leave()
  TEsound.stop('all')
  self.paused = false
end

-----------------------------------------------------------
-- Update functions
-----------------------------------------------------------
function Game:update(dt)
  if not self.paused then
    self:updateEntities(dt)
    self:updateCamera()
    self:removeDeadEntities()
  end
end

function Game:updateEntities(dt)
  for _,item in ipairs(self.entities) do
    if item['update'] then item:update(dt) end
  end
end

function Game:updateCamera()
  if self.camera and self.player then
    self:cameraFollowPlayer(dt)
  end
end

function Game:removeDeadEntities()
  for i=#self.entities,1,-1 do
    local item = self.entities[i]
    if item._remove then
      if self.world:hasItem(item) then
        self.world:remove(item)
      end
      table.remove(self.entities, i)
    end
  end
end

function Game:gameOver()
  self.playerKilled = true
  self.gameOverTimer = Timer.after(2, function()
    self.isGameOver = true
  end)
end

function Game:killedBoss()
  if self.bossIsDead == false then
    self.gameWinTimer = Timer.after(3, function()
      Gamestate.transition(Victory)
    end)
    self.bossIsDead = true
  end
end

-----------------------------------------------------------
-- General gameplay functions
-----------------------------------------------------------
function Game:screenShake(duration)
  local originalX, originalY = self.camera:position()

  local d = duration or .1
  Timer.during(d, function()
      self.camera:lookAt(originalX + math.random(-2,2), originalY + math.random(-2,2))
  end, function()
      -- reset camera position
      self.camera:lookAt(originalX, originalY)
  end)
end

function Game:cameraFollowPlayer()
  local settings = require 'setup.settings'
  local x, y = self.player.pos.x, self.player.pos.y
  local border = 96
  local xmin, xmax = border, settings.pxWidth - border
  local ymin, ymax = border, settings.pxHeight - border
  local smoother = self.cameraSmoother
  self.camera:lockWindow(x, y, xmin, xmax, ymin, ymax, smoother)
end

function Game:toggleMute()
  self.muted = not self.muted
  if self.muted then
    TEsound.volume('all', 0.0)
  else
    TEsound.volume('all', 1.0)
  end
end

function Game:togglePause()
  if not self.playerKilled and not self.isGameOver and not self.bossIsDead then
    self.paused = not self.paused
  end
end

-----------------------------------------------------------
-- Draw functions
-----------------------------------------------------------
function Game:draw()
  self.camera:attach()
  lg.clear({24, 20, 37})
  lg.setColor({255,255,255})

  table.sort(self.entities, function(a,b) return a.zIndex < b.zIndex end)

  for _,item in ipairs(self.entities) do
    if item['draw'] then item:draw() end
  end

  self.camera:detach()
  self:drawUI()
end

function Game:drawUI()
  -- SLOPPY CODE BELOW:

  if self.player and not self.playerKilled then
    local x, y = 16, 20
    local r = 0
    local sx, sy = 2, 2
    lg.print('HP: ' .. self.player.hp, x, 4, r, sx, sy)

    local rw = VeniceClassic:getWidth('HP: 000')
    local w = 240
    local h = 8 * sy
    local fw = (self.player.hp / 100) * 240

    lg.setLineWidth(2)
    lg.rectangle('line', (x + rw) * sx, y, w, h)
    lg.rectangle('fill', (x + rw) * sx, y, fw, h)
  end

  if self.isGameOver then
    local w, h = lg.getDimensions()
    local x, y = 0, h / 4
    local limit = w/2
    local r = 0
    local sx, sy = 2, 2
    local str = 'GAME OVER\n\nPress Enter to restart'

    -- draw box
    local textWidth, lines = VeniceClassic:getWrap(str, limit)
    local textHeight = #lines * VeniceClassic:getHeight()

    textWidth = (textWidth + 16) * sx
    textHeight = (textHeight + 16) * sy

    local textX = (w - textWidth) / 2

    lg.setColor({38,43,68})
    lg.rectangle('fill', textX, y, textWidth, textHeight)

    -- draw text
    lg.setColor({255,255,255})
    lg.printf(str, x, y, limit, 'center', r, sx, sy)
  end

  if self.isNewGame and not self.paused then
    local w, h = lg.getDimensions()
    local x, y = 0, 48
    local limit = w/2
    local r = 0
    local sx, sy = 2, 2
    local str = 'LONG LIVE THE KING'

    -- draw text
    lg.setColor({255,255,255})
    lg.printf(str, x, y, limit, 'center', r, sx, sy)
  end

  if self.bossIsDead and not self.paused then
    local w, h = lg.getDimensions()
    local x, y = 0, 48
    local limit = w/2
    local r = 0
    local sx, sy = 2, 2
    local str = 'WIZARD DEFEATED!'

    -- draw text
    lg.setColor({255,255,255})
    lg.printf(str, x, y, limit, 'center', r, sx, sy)
  end

  if self.paused then
    local w, h = lg.getDimensions()
    local x, y = 0, 48
    local limit = w/2
    local r = 0
    local sx, sy = 2, 2
    local str = 'PAUSED'

    -- draw text
    lg.setColor({255,255,255})
    lg.printf(str, x, y, limit, 'center', r, sx, sy)
  end

end


-----------------------------------------------------------
-- Input functions
-----------------------------------------------------------

function Game:keypressed(key, code)
  if key == 'return' or key == 'space' then
    if self.isGameOver then
      self:reset()
    end
  elseif key == 'm' then
    self:toggleMute()
  elseif key == 'p' then
    self:togglePause()
  end
end
